const std = @import("std");

const width = 64;
const height = 32;

var pixel_data = std.mem.zeroes([height * width * 3]u8);

export fn getPixelDataPointer() [*]u8 {
    return @ptrCast(&pixel_data);
}

const marker_color = Color{ .r = 0xAA, .g = 0x61, .b = 0xB7 };
const seconds_color = Color{ .r = 0xFF, .g = 0x98, .b = 0x00 };
const minutes_color = Color{ .r = 0x29, .g = 0xB6, .b = 0xF6 };
const hours_color = Color{ .r = 0x38, .g = 0x8E, .b = 0x3C };

const indicator_from = 6;
const indicator_to = 6;

const seconds_from = 5;
const seconds_to = 5;

const minutes_from = 0;
const minutes_to = 4;

const hours_from = 0;
const hours_to = 3;

const position = Vector2{ .x = 6, .y = 6 };

export fn draw() void {
    @memset(&pixel_data, 0);

    for (0..12) |a| {
        clockHand(position, indicator_from, indicator_to, (@as(f32, @floatFromInt(a)) / 12.0) * 360.0, marker_color);
    }

    const epoch_seconds = std.time.epoch.EpochSeconds{ .secs = timestamp() - std.time.s_per_hour * 6 };
    const day_seconds = epoch_seconds.getDaySeconds();

    const seconds: f32 = @floatFromInt(day_seconds.getSecondsIntoMinute());
    const minutes: f32 = @floatFromInt(day_seconds.getMinutesIntoHour());
    const hours: f32 = @floatFromInt(day_seconds.getHoursIntoDay());

    clockHand(position, seconds_from, seconds_to, 270 + (seconds / 60.0) * 360.0, seconds_color);
    clockHand(position, minutes_from, minutes_to, 270 + (minutes / 60.0) * 360.0, minutes_color);
    clockHand(position, hours_from, hours_to, 270 + (hours / 12.0) * 360.0, hours_color);
}

fn clockHand(pos: Vector2, from: u8, to: u8, angle: f32, color: Color) void {
    const radians = angle * std.math.pi / 180.0;
    const x: f32 = @floatFromInt(pos.x);
    const y: f32 = @floatFromInt(pos.y);
    const from_float: f32 = @floatFromInt(from);
    const to_float: f32 = @floatFromInt(to);
    line(
        .{
            .x = @intFromFloat(@round(x + @cos(radians) * from_float)),
            .y = @intFromFloat(@round(y + @sin(radians) * from_float)),
        },
        .{
            .x = @intFromFloat(@round(x + @cos(radians) * to_float)),
            .y = @intFromFloat(@round(y + @sin(radians) * to_float)),
        },
        color,
    );
}

fn line(from: Vector2, to: Vector2, color: Color) void {
    const dx = @abs(@as(i16, @intCast(to.x)) - @as(i16, @intCast(from.x)));
    const dy = @abs(@as(i16, @intCast(to.y)) - @as(i16, @intCast(from.y)));
    const sx: i2 = if (from.x < to.x) 1 else -1;
    const sy: i2 = if (from.y < to.y) 1 else -1;
    var v = from;

    if (dx > dy) {
        var err = @as(f32, @floatFromInt(dx)) / 2.0;
        while (v.x != to.x) {
            setPixel(v, color);
            err -= @floatFromInt(dy);
            if (err < 0) {
                v.y = @intCast(@as(i16, @intCast(v.y)) + sy);
                err += @floatFromInt(dx);
            }
            v.x = @intCast(@as(i16, @intCast(v.x)) + sx);
        }
    } else {
        var err = @as(f32, @floatFromInt(dy)) / 2.0;
        while (v.y != to.y) {
            setPixel(v, color);
            err -= @floatFromInt(dx);
            if (err < 0) {
                v.x = @intCast(@as(i16, @intCast(v.x)) + sx);
                err += @floatFromInt(dy);
            }
            v.y = @intCast(@as(i16, @intCast(v.y)) + sy);
        }
    }
    setPixel(v, color);
}

inline fn setPixel(pos: Vector2, color: Color) void {
    const ptr: *[height][width][3]u8 = @ptrCast(&pixel_data);
    ptr[pos.y][pos.x][0] = color.r;
    ptr[pos.y][pos.x][1] = color.g;
    ptr[pos.y][pos.x][2] = color.b;
}

inline fn getPixel(pos: Vector2) Color {
    const ptr: *[height][width][3]u8 = @ptrCast(&pixel_data);
    return .{
        .r = ptr[pos.x][pos.x][0],
        .g = ptr[pos.x][pos.x][1],
        .b = ptr[pos.x][pos.x][2],
    };
}

const Color = struct {
    r: u8,
    g: u8,
    b: u8,

    pub const WHITE = Color{ .r = 0xFF, .g = 0xFF, .b = 0xFF };
    pub const BLACK = Color{ .r = 0x00, .g = 0x00, .b = 0x00 };
};

const Vector2 = struct {
    x: u8,
    y: u8,
};

extern fn timestamp() u64;
extern fn printi64(arg: i64) void;
extern fn printf32(arg: f32) void;
